class TransactionHooksDatabaseWrapperMixin(object):
    """
    A ``DatabaseWrapper`` mixin to implement transaction-committed hooks.

    To use, create a package for your custom database backend and place a
    ``base.py`` module within it. Import whatever ``DatabaseWrapper`` you want
    to subclass (under some other name), and then create a ``DatabaseWrapper``
    class which inherits from both this mixin and the parent
    ``DatabaseWrapper`` (in that order).

    For an example, see ``backends/postgresql_psycopg2/base.py``.

    """
    def __init__(self, *a, **kw):
        super(TransactionHooksDatabaseWrapperMixin, self).__init__(*a, **kw)

        # a list of no-argument functions to run when the transaction commits;
        # each entry is an (sids, func) tuple, where sids is a list of the
        # active savepoint IDs when this function was registered
        self.run_on_commit = []


    def on_commit(self, func):
        if self.in_atomic_block:
            # transaction in progress; save for execution on commit
            self.run_on_commit.append((self.savepoint_ids[:], func))
        else:
            # no transaction in progress; execute immediately
            func()


    def run_and_clear_commit_hooks(self):
        for sids, func in self.run_on_commit:
            func()

        self.run_on_commit = []


    def commit(self, *a, **kw):
        super(TransactionHooksDatabaseWrapperMixin, self).commit(*a, **kw)

        self.run_and_clear_commit_hooks()


    def savepoint_rollback(self, sid, *a, **kw):
        super(TransactionHooksDatabaseWrapperMixin, self).savepoint_rollback(
            sid, *a, **kw)

        # remove any callbacks registered while this savepoint was active
        self.run_on_commit = list(filter(
            lambda x: sid not in x[0], self.run_on_commit))


    def rollback(self, *a, **kw):
        super(TransactionHooksDatabaseWrapperMixin, self).rollback(*a, **kw)

        self.run_on_commit = []


    def connect(self, *a, **kw):
        super(TransactionHooksDatabaseWrapperMixin, self).connect(*a, **kw)

        self.run_on_commit = []


    def close(self, *a, **kw):
        super(TransactionHooksDatabaseWrapperMixin, self).close(*a, **kw)

        self.run_on_commit = []



from django.db.backends.mysql import base

from transaction_hooks.mixin import TransactionHooksDatabaseWrapperMixin


class DatabaseWrapper(TransactionHooksDatabaseWrapperMixin,
                      base.DatabaseWrapper):
    pass



from django.db import models
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class Thing(models.Model):
    num = models.IntegerField()


    def __str__(self):
        return "Thing %d" % self.num



import django
from django.db import connection
from django.db.transaction import atomic
import pytest

from .models import Thing


# Make sure Django app registry is ready (in 1.7+); workaround for
# pytest-django not being updated to 1.7+ yet.
@pytest.mark.tryfirst
@pytest.fixture(scope='session', autouse=True)
def setup_django():
    setup = getattr(django, 'setup', lambda: None)
    setup()



class Tracker(object):
    """Simulate the pattern of creating a DB object and notifying about it."""
    def __init__(self):
        self.notified = []


    def notify(self, id_):
        self.notified.append(id_)


    def do(self, num):
        """Create a Thing instance and notify about it."""
        Thing.objects.create(num=num)
        connection.on_commit(lambda: self.notify(num))


    def assert_done(self, nums):
        self.assert_notified(nums)
        assert sorted(t.num for t in Thing.objects.all()) == sorted(nums)


    def assert_notified(self, nums):
        assert self.notified == nums



@pytest.fixture
def track():
    """Return a new ``Tracker`` instance."""
    return Tracker()



class ForcedError(Exception):
    pass



@pytest.mark.usefixtures('transactional_db')
class TestConnectionOnCommit(object):
    """
    Tests for connection.on_commit().

    Creation/checking of database objects in parallel with callback tracking is
    to verify that the behavior of the two match in all tested cases.

    """
    def test_executes_immediately_if_no_transaction(self, track):
        track.do(1)
        track.assert_done([1])


    def test_delays_execution_until_after_transaction_commit(self, track):
        with atomic():
            track.do(1)
            track.assert_notified([])
        track.assert_done([1])


    def test_does_not_execute_if_transaction_rolled_back(self, track):
        try:
            with atomic():
                track.do(1)
                raise ForcedError()
        except ForcedError:
            pass

        track.assert_done([])


    def test_executes_only_after_final_transaction_committed(self, track):
        with atomic():
            with atomic():
                track.do(1)
                track.assert_notified([])
            track.assert_notified([])
        track.assert_done([1])


    def test_discards_hooks_from_rolled_back_savepoint(self, track):
        with atomic():
            # one successful savepoint
            with atomic():
                track.do(1)
            # one failed savepoint
            try:
                with atomic():
                    track.do(2)
                    raise ForcedError()
            except ForcedError:
                pass
            # another successful savepoint
            with atomic():
                track.do(3)

        # only hooks registered during successful savepoints execute
        track.assert_done([1, 3])


    def test_no_hooks_run_from_failed_transaction(self, track):
        """If outer transaction fails, no hooks from within it run."""
        try:
            with atomic():
                with atomic():
                    track.do(1)
                raise ForcedError()
        except ForcedError:
            pass

        track.assert_done([])


    def test_inner_savepoint_rolled_back_with_outer(self, track):
        with atomic():
            try:
                with atomic():
                    with atomic():
                        track.do(1)
                    raise ForcedError()
            except ForcedError:
                pass
            track.do(2)

        track.assert_done([2])


    def test_no_savepoints_atomic_merged_with_outer(self, track):
        with atomic():
            with atomic():
                track.do(1)
                try:
                    with atomic(savepoint=False):
                        raise ForcedError()
                except ForcedError:
                    pass

        track.assert_done([])


    def test_inner_savepoint_does_not_affect_outer(self, track):
        with atomic():
            with atomic():
                track.do(1)
                try:
                    with atomic():
                        raise ForcedError()
                except ForcedError:
                    pass

        track.assert_done([1])


    def test_runs_hooks_in_order_registered(self, track):
        with atomic():
            track.do(1)
            with atomic():
                track.do(2)
            track.do(3)

        track.assert_done([1, 2, 3])


    def test_hooks_cleared_after_successful_commit(self, track):
        with atomic():
            track.do(1)
        with atomic():
            track.do(2)

        track.assert_done([1, 2]) # not [1, 1, 2]


    def test_hooks_cleared_after_rollback(self, track):
        try:
            with atomic():
                track.do(1)
                raise ForcedError()
        except ForcedError:
            pass

        with atomic():
            track.do(2)

        track.assert_done([2])


    def test_hooks_cleared_on_reconnect(self, track):
        with atomic():
            track.do(1)
            connection.close()

        connection.connect()

        with atomic():
            track.do(2)

        track.assert_done([2])



from .settings import *

import os

DATABASES = {
    'default': {
        'ENGINE': 'transaction_hooks.backends.sqlite3',
        'TEST_NAME': os.path.join(
            os.path.dirname(os.path.abspath('__file__')),
            'testdb.sqlite',
            )
        },
    }



from os.path import join
from setuptools import setup, find_packages


long_description = (
    open('README.rst').read() + open('CHANGES.rst').read())


def get_version():
    with open(join('transaction_hooks', '__init__.py')) as f:
        for line in f:
            if line.startswith('__version__ ='):
                return line.split('=')[1].strip().strip('"\'')


setup(
    name='django-transaction-hooks',
    version=get_version(),
    description=(
        'Django database backends that permit registering '
        'post-transaction-commit hooks'
        ),
    long_description=long_description,
    author='Carl Meyer',
    author_email='carl@oddbird.net',
    url='https://github.com/carljm/django-transaction-hooks/',
    packages=find_packages(),
    install_requires=['django>=1.6.1'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Framework :: Django',
    ],
    zip_safe=False,
)
